<?php

namespace Controllers;

use Controllers\ReportWorkflowController as Report;
use Services\AprovacaoAlcadaService as Service;
use Controllers\SupplyTekApprovalWorkflowController as SupplyTekApproval;
use Zeedhi\Framework\DTO\Request\Row;
use Zeedhi\Framework\DTO\Response;
use Zeedhi\Framework\DTO\Response\Error;

class AprovacaoAlcadasController
{
    private $totalAFAtual = 0;
    private $valorDemanda = 0;
    private $valorAprovacaoAtual;
    private $supplyTekApproval;
    
    public function __construct(Service $service, Report $report, SupplyTekApproval $supplyTekApproval)
    {
        $this->service = $service;
        $this->report = $report;
        $this->supplyTekApproval = $supplyTekApproval;
    }

    public function getRequestToAprrove(Row $request, Response $response)
    {
        try {
            $row = $request->getRow();
            $idSisDos = $this->service->getIdSisDos();
            $order = $this->service->getApproverOrder($row);
            
            if (isset($order[0]) && $order[0]["IDEXIGETAPROVE"] == "S" || $idSisDos[0]["IDSISDOS"] == "3") {
                $rowToSend = self::arrayWorkflow($row);
                $this->supplyTekApproval->startWorkflow($rowToSend, $response);
            } else {
                if ($row['OPERACAO'] == 'APROVAR') {
                    self::orderApprove($row);
                } else if ($row['OPERACAO'] == 'REPROVAR') {
                    self::orderRepprove($row);
                }
            }
            if ($row['OPERACAO'] == 'APROVAR') {
                $messageToResponse = "SUCCESS_APPROVED";
            } else if ($row['OPERACAO'] == 'REPROVAR') {
                $messageToResponse = "SUCCESS_REPROVED";
            }
            
            $this->updateOrca($row);
            
            $messageToResponse = new Response\Message($messageToResponse);
            $response->addMessage($messageToResponse);
        } catch (\Exception $e) {
            $response->setError(new Error($e->getMessage(), 500));
        }
    }
    
    public function orderApprove($approvalData)
    {
        $orderTag = 'APROVADO';
        $orderTagFinalize = 'APROVADA';
        $justifyMessage = $approvalData['JUSTIFICATIVA'] == '' ? 'Aprovado.' : $approvalData['JUSTIFICATIVA'];
        $orderData = $this->service->getOrder($approvalData);
        $usaPorcentagem  = ($orderData['IDTIPOVRAPROV'] == 'PERCENTUAL');
        $paramValorUtilizado = self::getParamConsisteValorAprovacao($orderData);
        $maxValue = self::getApproversMaxValue($orderData, $usaPorcentagem);
        $totalOrderValue = self::getTotal($orderData, $paramValorUtilizado, $approvalData, $usaPorcentagem);
        $approverMaxValue = self::getTotalDisponivelParaUso();
        if ($totalOrderValue > $maxValue) {
            if ($paramValorUtilizado == 'M') {
                throw new \Exception('N_VALUE_APPROVED_TOTAL_AF_PERIOD');
            }
            
            throw new \Exception('N_VALUE_APPROVED_TOTAL_AF');
        }
        
        /*if($totalOrderValue > $approverMaxValue){
            throw new \Exception('N_VALUE_ENOUGH_APPROVED_AF');
        }*/
        
        // DEBUG - Comentar a linha abaixo para que não seja realizado o update.
        self::updateApprovalWorkflow($approvalData, $orderData, $justifyMessage, $orderTag);
        $subjectParam = self::setAprroveData($orderData, $justifyMessage, $orderTagFinalize, $approvalData, $totalOrderValue);
        $this->report->sendApprovalReport($orderData, $subjectParam, $justifyMessage, $approvalData['CDOPERLOG']);
    }

    public function orderRepprove($approvalData)
    {
        $orderTagUpdate = 'RECUSADO';
        $orderTagFinalize = 'RECUSADA';
        $subjectParam = 'REPROVACAO_ALCADA';
        $justifyMessage = $approvalData['JUSTIFICATIVA'];
        $orderData = $this->service->getOrder($approvalData);
        
        if (!empty($justifyMessage)) {
            self::updateApprovalWorkflow($approvalData, $orderData, $justifyMessage, $orderTagUpdate);
            self::finalizeWorkflow($orderData, $orderTagFinalize);
            $this->report->sendRepproveReport($orderData, $subjectParam, $justifyMessage, $approvalData['CDOPERLOG']);
        }
    }

    private function updateApprovalWorkflow($approvalData, $orderData, $justifyMessage, $orderTag)
    {
        $userLog = $approvalData['CDOPERLOG'];
        $approveCode = $orderData['NRAPROVALCADA'];
        $this->service->updateWorkflow($approveCode, $orderTag, $justifyMessage, $userLog);
    }

    private function finalizeWorkflow($orderData, $orderTag)
    {
        $nrAutoForn = $orderData['NRAUTOFORN'];
        $cdSupplier = $orderData['CDFORNECED'];
        $this->service->updateOrder($orderData, $orderTag);
        $this->service->executeCONSISTE_ORCAMENTO($orderData['CDFILDEST'], $orderData['CDFILLANC'], $orderData['NRAUTOFORN']);
    }

    private function setApprovalOnWorkflow($orderData, $userCode, $justifyMessage)
    {
        $nrApprove = $orderData['NRAPROVALCADA'];
        $this->service->setApprover($userCode, $nrApprove);
    }

    private function setAprroveData($orderData, $justifyMessage, $orderTag, $approvalData, $totalOrderValue)
    {
        $approver = $this->service->getApprover($orderData);
        
        if (empty($approver)) {
            self::finalizeWorkflow($orderData, $orderTag);
            
            return 'LIBERACAO_DEMANDA';
        }
        
        if ($this->isValorExcedente()) {
            $userCode = $approver[0]['CDOPERADOR'];
            self::setApprovalOnWorkflow($orderData, $userCode, $justifyMessage);
            
            return 'APROVACAO_ALCADA';
        }
        
        self::finalizeWorkflow($orderData, $orderTag);
        
        return 'APROVACAO_AUTOMATICA';
    }
    
    /**
     * Verifica se o total disponível ultrapassa o permitido, sendo necessário
     * ir para o próximo aprovador.
     * 
     * Se o valor for positivo, significa que o valor ultrapassou o total disponível e,
     * portanto, deve ir para o próximo aprovador.
     * Do contrário, o valor está dentro do permitido e é aprovado automaticamente.
     */
    private function isValorExcedente()
    {
        $totalDisponivel = $this->getTotalDisponivel();
        
        return $totalDisponivel > 0;
    }
    
    /**
     * Retorna o total disponível, calculado entre Total Gasto menos o Total Disponível para Uso.
     * 
     * Os valores que compõe os valores utilizados no cálculo são recuperados
     * na função getTotalGastoNoPeriodo.
     */
    private function getTotalDisponivel()
    {
        return $this->getTotalGastoNoPeriodo() - $this->getTotalDisponivelParaUso();
    }
    
    /**
     * O total gasto é a soma dos seguintes valores:
     * - resultado da função VALOR_DEMANDA (detalhes do funcionamento no service);
     * - valor total da AF atual.
     */
    private function getTotalGastoNoPeriodo()
    {
        // return $this->totalPeriodo + $this->totalAFAtual;
        return $this->valorDemanda + $this->totalAFAtual;
    }
    
    /**
     * Total disponível para uso é o valor de aprovação do aprovador atual.
     * O valor do orçamento disponível já foi descontado na VALOR_DEMANDA.
     */
    private function getTotalDisponivelParaUso()
    {
        // return $this->totalOrcamento + $this->valorAprovacaoAtual;
        return $this->valorAprovacaoAtual;
    }
    
    public function getTotal($dadosDemanda, $paramValorUtilizado, $approvalData, $usaPorcentagem = false)
    {
        $aprovadorAtual = $this->service->findAprovadorAtual($approvalData)[0];
        
        if($usaPorcentagem) {
            $aprovadorAtual['VRAPROVCMP'] = self::buscaValorTotalPorcentagem($dadosDemanda, $aprovadorAtual['VRAPROVCMP']);
        }
        
        $this->totalAFAtual = $dadosDemanda['VRTOTALAF'];
        
        if ($paramValorUtilizado == 'M') {
            $this->valorAprovacaoAtual = $aprovadorAtual['VRAPROVCMP'];
            $dataEntregaAF    = $this->getDataEntregaAF($dadosDemanda);
            $periodoOrcamento = $this->getPeriodoOrcamento($dadosDemanda, $dataEntregaAF);
            $this->valorDemanda = $this->getValorDemanda($dadosDemanda, $dataEntregaAF, $periodoOrcamento);
            $totalGasto = $this->getTotalGastoNoPeriodo();
        } else {
            $orcamento = self::buscaOrcamento($dadosDemanda);
            $this->valorDemanda = 0;
            $this->valorAprovacaoAtual = $orcamento + $aprovadorAtual['VRAPROVCMP'];
            $totalGasto = $dadosDemanda['VRTOTALAF'];
            //$totalGasto = self::getTotalDia($dadosDemanda) + $dadosDemanda['VRTOTALAF']; 
        }
        return $totalGasto;
    }
    
    public function getValorDemanda($dadosDemanda, $dataEntregaAF, $periodoOrcamento)
    {
        $paramsValorDemanda = [
            'DTPROGENTAF' => $dataEntregaAF,
            'CDFILDEST'   => $dadosDemanda['CDFILDEST'],
            'NRDEMANDA'   => $dadosDemanda['NRAUTOFORN'],
            'IDPERIODOORCA'   => $periodoOrcamento['V_IDPERIODOORCA'],
            'NRPERIODOORCA'   => $periodoOrcamento['V_NRPERIODOORCAGOP'],
        ];
        
        return $this->service->getValorDemanda($paramsValorDemanda)[0]['VRDEMANDA'];
    }
    
    public function getTotalDia($dadosDemanda)
    {
        $params = [
            "CDFILDEST"    => $dadosDemanda['CDFILDEST'],
            "CDFILLANC"    => $dadosDemanda['CDFILLANC'],
            "NRDEMANDA"    => $dadosDemanda['NRAUTOFORN'],
        ];
        
        return $this->service->totalDiaAF($params)[0]['TOTAL'];
    }
    
    /*
    // Descontinuado
    public function getProdutoOrcamento($dadosDemanda, $periodoOrcamento, $dataEntregaAF)
    {
        $paramsfindProdutoOrcamento = [
            'IDPERIODOORCA' => $periodoOrcamento['V_IDPERIODOORCA'],
            'NRPERIODOORCA' => $periodoOrcamento['V_NRPERIODOORCAGOP'],
            'DTPROGENTAF'   => $dataEntregaAF,
            'CDFILDEST'     => $dadosDemanda['CDFILDEST'],
            'CDFILLANC'     => $dadosDemanda['CDFILLANC'],
            'NRDEMANDA'     => $dadosDemanda['NRAUTOFORN'],
        ];
            
        return $this->service->findProdutoOrcamento($paramsfindProdutoOrcamento);
    }*/
    
    public function getDiaInicioOrcamento($dadosDemanda)
    {
        return $this->service->findDiaInicioOrcamento($dadosDemanda['CDFILDEST'])[0]['DIAINIORCA'];
    }
    
    public function getPeriodoOrcamento($dadosDemanda, $dataEntregaAF)
    {
        $diaInicioOrcamento = $this->getDiaInicioOrcamento($dadosDemanda);
        
        $dataOrcamento = $this->getDataOrcamento($diaInicioOrcamento, $dataEntregaAF);
        
        $paramsPeriodoOrcamento = [
            'V_DATAINI' => $dataOrcamento['V_DATAINI'],
            'DIAINIORCA' => $diaInicioOrcamento,
        ];
        
        return $this->service->findPeriodoOrcamento($paramsPeriodoOrcamento)[0];
    }
    
    public function getDataOrcamento($diaInicioOrcamento, $dataEntregaAF)
    {
        $params = [
            'DIA' => $diaInicioOrcamento,
            'DATA' => $dataEntregaAF,
        ];
        
        return $this->service->findDataOrcamento($params)[0];
    }
    
    public function getDataEntregaAF($dadosDemanda)
    {
        $params = [
            'CDFILDEST' => $dadosDemanda['CDFILDEST'],
            'CDFILLANC' => $dadosDemanda['CDFILLANC'],
            'NRDEMANDA' => $dadosDemanda['NRAUTOFORN'],
        ];
        
        return $this->service->findDataEntregaAF($params)[0]['DTPROGENTAF'];
    }
    
    public function getParamConsisteValorAprovacao($dadosDemanda)
    {
        $params = [
            'CDFILDEST'      => $dadosDemanda['CDFILDEST'],
            'NRCONFIGALCADA' => $dadosDemanda['NRCONFIGALCADA'],
        ];
        
        return $this->service->findFilialAlcada($params)[0]['IDCONSISTVRAPROV'];
    }
    
    public function arrayWorkflow($row)
    {
        $row["JUSTIFICATIVA"] = $row['JUSTIFICATIVA'] == '' ? 'Aprovado.' : $row['JUSTIFICATIVA'];
        
        $rowToSend = array(
            "CDFILDEST"         => $row["CDFILDEST"],
            "CDFILLANC"         => $row["CDFILLANC"],
            "NRAUTOFORN"        => $row["NRAUTOFORNE"],
            "CDOPERADOR"        => $row["CDOPERLOG"],
            "STATUS"            => $row["OPERACAO"][0],
            "DSJUSTIFIC"        => $row["JUSTIFICATIVA"],
            "NRAPROVALCADA"     => $row["NRAPROVALCADA"],
            "NRCONFIGALCADA"    => $row["NRCONFIGALCADA"],
            "VRAPROVCMP"        => $row["VRAPROVACAO"]
        );
        
        return $rowToSend;
    }
        
    public function buscaValorTotalPorcentagem($dadosDemanda, $porcentagemAprovador)
    {
        $orcamento = self::buscaOrcamento($dadosDemanda);
        $vraprovcmp = $orcamento * $porcentagemAprovador / 100;
        return $vraprovcmp;
    }
    
    public function buscaOrcamento ($dadosDemanda)
    {
        $dadosOrcamento['CDFILDEST']  = $dadosDemanda['CDFILDEST'];
        $dadosOrcamento['CDFILLANC']  = $dadosDemanda['CDFILLANC'];
        $dadosOrcamento['NRAUTOFORN'] = $dadosDemanda['NRAUTOFORN'];
        
        $diainiorca = $this->service->buscaDIAINIORCA($dadosOrcamento);
        $dadosOrcamento['DIAINIORCA'] = $diainiorca[0]['DIAINIORCA'];
        
        $dtprogentaf = $this->service->buscaDTPROGENTAF($dadosOrcamento);
        $dadosOrcamento['DTPROGENTAF'] = substr($dtprogentaf[0]['DTPROGENTAF'],0,10);
        
        $datacase = self::createDATACASE($dadosOrcamento['DIAINIORCA'], $dadosOrcamento['DTPROGENTAF']);
        $dadosOrcamento['DATACASE'] = $datacase;
        

        $dataini = $this->service->buscaDATAINI($dadosOrcamento);
        $dadosOrcamento['DATAINI'] = substr($dataini[0]['DATAINI'],0,10);
        

        $datafim = $this->service->createDATAFIM($dadosOrcamento);
        $dadosOrcamento['DATAFIM'] = substr($datafim[0]['DATAFIM'],0,10);
        

        $periodo = $this->service->buscaPeriodo($dadosOrcamento);
        $dadosOrcamento['NRPERIODO'] = $periodo[0]['MAX_NRPERIODO'];
        $dadosOrcamento['IDPERIODOORCA'] = $periodo[0]['MAX_IDPERIODOORCA'];


        $orcamento = $this->service->buscaOrcamento($dadosOrcamento);
        
        if($orcamento[0]['VRORCAITEM'] == 0){
            throw new \Exception('N_VALUE_AF_SELECT');
         }
        
        return $orcamento[0]['VRORCAITEM'];
    }
    
    public function createDATACASE ($diainiorca,$dtprogentaf)
    {
        $mes = substr($dtprogentaf,3,2);
        $ano = substr($dtprogentaf,6,4);
        $ultimo_dia = date("t", mktime(0,0,0,$mes,'01',$ano));
        $resultado = $diainiorca > $ultimo_dia ? $ultimo_dia : $resultado = substr($diainiorca,0,2);
        $resultado = strlen($resultado) == 1 ? $resultado = '0'.$resultado : $resultado;
        
        return $resultado;
    }
    
    public function getApproversMaxValue($orderData, $usaPorcentagem)
    {
        $maxvalue = $this->service->getApproversMaxValue($orderData)['VALORMAXIMO'];
        if($usaPorcentagem){
            $orcamento = self::buscaOrcamento($orderData);
            $maxvalueApprover = $orcamento * $maxvalue / 100;
            $maxvalue = $maxvalueApprover + $orcamento;
        }
        return $maxvalue;
    }
    
    private function updateOrca($row)
    {
        $autoforne = $this->service->getOrder($row);
        
        $params = [
            'CDFILDEST' => $autoforne['CDFILDEST'],
            'CDFILLANC' => $autoforne['CDFILLANC'],
            'NRDEMANDA' => $autoforne['NRAUTOFORN'],
        ];
        
        $dtEntregas = $this->service->findDataEntregaAF($params);
        
        foreach ($dtEntregas as $dtEntr) {
            //Execução para atualizar o orçamento com a demanda atual aprovada
            $this->service->executeCONSISTE_ORCAMENTO($autoforne['CDFILDEST'], $autoforne['CDFILLANC'], null, $dtEntr['DTPROGENTAF']);
        }
    }
}